//  ParticleSystem.cpp
//  Description: Particle system class

#include "ParticleSystem.h"
#include <math.h>
#include <iostream>
#include "FlyCamera.h"
#include "DrawingState.h"

// OpenGL includes
#include <Fl/Gl.h>
#include <Gl/Glu.h>

#define SKY_WIDTH 400
#define SKY_HEIGHT 100

Particle::Particle()
{
}

ParticleSystem::ParticleSystem(int mParticles, float height)
{
	maxParticles = mParticles;
	emissionHeight = height;
	animate = false;
}

ParticleSystem::~ParticleSystem()
{
}

void ParticleSystem::draw(DrawingState * d)
{
	static float SNOW_SIZE = .5f;
	int i;
	int count = (int)particles.size();
	Particle p;
	FlyCamera * cam = (FlyCamera *)d->camera;
	float rotY;
	float pos[3];

	if (d->timeOfDay == 15) {
		if (!animate)
			this->emit();
		animate = true;
	} else {
		if (animate) {
			animate = false;
			particles.clear();
		}
		return;
	}

	if (!show_shadow) glColor3f(1, 1, 1);
	for (i = 0; i < count; i++) {
		p = (Particle)particles[i];

		pos[0] = p.pos[0] + transform[3][0] - cam->posX;
		pos[1] = p.pos[1] + transform[3][1] - cam->posY;
		pos[2] = p.pos[2] + transform[3][2] - cam->posZ;

		rotY = -(cam->direction + atan2(pos[2], pos[0]) * 180.f / 3.14159f);
		glPushMatrix();
			glTranslatef(p.pos[0], p.pos[1], p.pos[2]);
			glRotatef(rotY, 0, 1, 0);
			glBegin(GL_TRIANGLES);
				glNormal3f(0, 1, 0);
				glVertex3f(0, 0, -(SNOW_SIZE / 2.f));
				glVertex3f(0, 0, (SNOW_SIZE / 2.f));
				glVertex3f(0, SNOW_SIZE, 0);
				glVertex3f(0, 0, -(SNOW_SIZE / 2.f));
				glVertex3f(0, -SNOW_SIZE, 0);
				glVertex3f(0, 0, (SNOW_SIZE / 2.f));
			glEnd();
		glPopMatrix();
	}
}

void ParticleSystem::update(unsigned long time)
{
	int i, component;
	int count = (int)particles.size();
	Particle p;
	float amtTime = (float)time / 100.f;

	if (!animate) return;

	this->emit();

	// Move particles
	for (i = 0; i < count; i++) {
		p = (Particle)particles[i];

		for (component = 0; component < 3; component++) {
			p.pos[component] += (p.vel[component] * amtTime);
			p.vel[component] += (p.accel[component] * amtTime);
			p.accel[component] += (amtTime * (float)((rand() % 100) - 50) / 10000);
		}

		//if (p.pos[0] < 0) p.pos[0] = 0;
		//if (p.pos[0] > SKY_WIDTH) p.pos[0] = SKY_WIDTH;
		//if (p.pos[2] < 0) p.pos[2] = 0;
		//if (p.pos[2] > SKY_HEIGHT) p.pos[2] = SKY_HEIGHT;

		if (p.accel[1] > 0)
			p.accel[1] = 0;

		if (p.vel[1] > 0)
			p.vel[1] = 0;

		particles[i] = p;

		if (p.pos[1] <= 0) {
			particles.erase(particles.begin() + i, particles.begin() + i + 1);
			i--;
			count--;
		}
	}
}

void ParticleSystem::emit(void)
{
	int freeParticles = maxParticles - (int)particles.size();
	int newParticles = (rand() % freeParticles);
	int i;

	for (i = 0; i < newParticles; i++) {
		this->addParticle();
	}
}

void ParticleSystem::addParticle(void)
{
	Particle p;

	p.pos[0] = (float)(rand() % SKY_WIDTH);
	p.pos[1] = emissionHeight;
	p.pos[2] = (float)(rand() % SKY_HEIGHT);
	p.size = (float)(rand() % 100) / 150;
	p.vel[0] = (float)((rand() % 100) - 50) / 200;
	p.vel[1] = -p.size; // Gravity
	p.vel[2] = (float)((rand() % 100) - 50) / 200;
	p.accel[0] = 0;
	p.accel[1] = 0;
	p.accel[2] = 0;

	particles.push_back(p);
}